@@ -1,4 +1,5 @@
 /* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
+/* vim: set ts=2 sw=2 et tw=78: */
 /* ***** BEGIN LICENSE BLOCK *****
  * Version: MPL 1.1/GPL 2.0/LGPL 2.1
  *
@@ -57,6 +58,7 @@
 #include "nsIDOMWindow.h"
 #include "nsIDOMChromeWindow.h"
 #include "nsIDOMWindowInternal.h"
+#include "nsIScriptObjectPrincipal.h"
 #include "nsIScreen.h"
 #include "nsIScreenManager.h"
 #include "nsIScriptContext.h"
@@ -81,6 +83,7 @@
 #include "nsIMarkupDocumentViewer.h"
 #include "nsIContentViewer.h"
 #include "nsIDocumentViewer.h"
+#include "nsIWindowProvider.h"
 
 #include "nsIPrefBranch.h"
 #include "nsIPrefService.h"
@@ -377,26 +380,22 @@ JSContextAutoPopper::~JSContextAutoPopper()
 
 nsresult JSContextAutoPopper::Push(JSContext *cx)
 {
-  nsresult rv = NS_OK;
-
   if (mContext) // only once
     return NS_ERROR_FAILURE;
 
   mService = do_GetService(sJSStackContractID);
   if(mService) {
-    if (cx) {
-      mContext = cx;
-    } else {
-      rv = mService->GetSafeJSContext(&mContext);
+    // Get the safe context if we're not provided one.
+    if (!cx && NS_FAILED(mService->GetSafeJSContext(&cx))) {
+      cx = nsnull;
     }
 
-    if (NS_SUCCEEDED(rv) && mContext) {
-      rv = mService->Push(mContext);
-      if (NS_FAILED(rv))
-        mContext = 0;
+    // Save cx in mContext to indicate need to pop.
+    if (cx && NS_SUCCEEDED(mService->Push(cx))) {
+      mContext = cx;
     }
   }
-  return mContext ? NS_OK : NS_ERROR_FAILURE; 
+  return mContext ? NS_OK : NS_ERROR_FAILURE;
 }
 
 /****************************************************************
@@ -465,13 +464,18 @@ nsWindowWatcher::OpenWindow(nsIDOMWindow *aParent,
   jsval    *argv = nsnull;
   JSContext *cx;
   void *mark;
-  nsresult  rv;
 
-  rv = ConvertSupportsTojsvals(aParent, aArguments, &argc, &argv, &cx, &mark);
+  // This kungFuDeathGrip is filled when we are using aParent's context. It
+  // prevents the context from being destroyed before we're truly done with
+  // it.
+  nsCOMPtr<nsIScriptContext> kungFuDeathGrip;
+
+  nsresult rv = ConvertSupportsTojsvals(aParent, aArguments, &argc, &argv, &cx,
+                                        &mark, getter_AddRefs(kungFuDeathGrip));
   if (NS_SUCCEEDED(rv)) {
     PRBool dialog = argc == 0 ? PR_FALSE : PR_TRUE;
-    rv = OpenWindowJS(aParent, aUrl, aName, aFeatures, dialog, argc, argv,
-                      _retval);
+    rv = OpenWindowJSInternal(aParent, aUrl, aName, aFeatures, dialog, argc,
+                              argv, PR_FALSE, _retval);
 
     if (argv) {
       js_FreeStack(cx, mark);
@@ -481,6 +485,47 @@ nsWindowWatcher::OpenWindow(nsIDOMWindow *aParent,
   return rv;
 }
 
+struct SizeSpec {
+  SizeSpec() :
+    mLeftSpecified(PR_FALSE),
+    mTopSpecified(PR_FALSE),
+    mOuterWidthSpecified(PR_FALSE),
+    mOuterHeightSpecified(PR_FALSE),
+    mInnerWidthSpecified(PR_FALSE),
+    mInnerHeightSpecified(PR_FALSE),
+    mUseDefaultWidth(PR_FALSE),
+    mUseDefaultHeight(PR_FALSE)
+  {}
+  
+  PRInt32 mLeft;
+  PRInt32 mTop;
+  PRInt32 mOuterWidth;  // Total window width
+  PRInt32 mOuterHeight; // Total window height
+  PRInt32 mInnerWidth;  // Content area width
+  PRInt32 mInnerHeight; // Content area height
+
+  PRPackedBool mLeftSpecified;
+  PRPackedBool mTopSpecified;
+  PRPackedBool mOuterWidthSpecified;
+  PRPackedBool mOuterHeightSpecified;
+  PRPackedBool mInnerWidthSpecified;
+  PRPackedBool mInnerHeightSpecified;
+
+  // If these booleans are true, don't look at the corresponding width values
+  // even if they're specified -- they'll be bogus
+  PRPackedBool mUseDefaultWidth;
+  PRPackedBool mUseDefaultHeight;
+
+  PRBool PositionSpecified() const {
+    return mLeftSpecified || mTopSpecified;
+  }
+  
+  PRBool SizeSpecified() const {
+    return mOuterWidthSpecified || mOuterHeightSpecified ||
+      mInnerWidthSpecified || mInnerHeightSpecified;
+  }
+};
+
 NS_IMETHODIMP
 nsWindowWatcher::OpenWindowJS(nsIDOMWindow *aParent,
                               const char *aUrl,
@@ -490,11 +535,28 @@ nsWindowWatcher::OpenWindowJS(nsIDOMWindow *aParent,
                               PRUint32 argc,
                               jsval *argv,
                               nsIDOMWindow **_retval)
+{
+  return OpenWindowJSInternal(aParent, aUrl, aName, aFeatures, aDialog, argc,
+                              argv, PR_TRUE, _retval);
+}
+
+nsresult
+nsWindowWatcher::OpenWindowJSInternal(nsIDOMWindow *aParent,
+                                      const char *aUrl,
+                                      const char *aName,
+                                      const char *aFeatures,
+                                      PRBool aDialog,
+                                      PRUint32 argc,
+                                      jsval *argv,
+                                      PRBool aCalledFromJS,
+                                      nsIDOMWindow **_retval)
 {
   nsresult                        rv = NS_OK;
   PRBool                          nameSpecified,
                                   featuresSpecified,
+                                  isNewToplevelWindow = PR_FALSE,
                                   windowIsNew = PR_FALSE,
+                                  windowNeedsName = PR_FALSE,
                                   windowIsModal = PR_FALSE,
                                   uriToLoadIsChrome = PR_FALSE;
   PRUint32                        chromeFlags;
@@ -509,8 +571,7 @@ nsWindowWatcher::OpenWindowJS(nsIDOMWindow *aParent,
   NS_ENSURE_ARG_POINTER(_retval);
   *_retval = 0;
 
-  if (aParent)
-    GetWindowTreeOwner(aParent, getter_AddRefs(parentTreeOwner));
+  GetWindowTreeOwner(aParent, getter_AddRefs(parentTreeOwner));
 
   if (aUrl) {
     rv = URIfromURL(aUrl, aParent, getter_AddRefs(uriToLoad));
@@ -522,7 +583,9 @@ nsWindowWatcher::OpenWindowJS(nsIDOMWindow *aParent,
   nameSpecified = PR_FALSE;
   if (aName) {
     CopyUTF8toUTF16(aName, name);
+#ifdef DEBUG
     CheckWindowName(name);
+#endif
     nameSpecified = PR_TRUE;
   }
 
@@ -534,39 +597,9 @@ nsWindowWatcher::OpenWindowJS(nsIDOMWindow *aParent,
   }
 
   // try to find an extant window with the given name
-  if (nameSpecified) {
-    nsCOMPtr<nsIJSContextStack> stack =
-      do_GetService(sJSStackContractID);
-
-    JSContext *cx = nsnull;
-
-    if (stack) {
-      stack->Peek(&cx);
-    }
-
-    nsCOMPtr<nsIDocShellTreeItem> callerItem;
-
-    if (cx) {
-      nsCOMPtr<nsIWebNavigation> callerWebNav =
-        do_GetInterface(nsWWJSUtils::GetDynamicScriptGlobal(cx));
-
-      callerItem = do_QueryInterface(callerWebNav);
-    }
-
-    nsCOMPtr<nsIDocShellTreeItem> parentItem;
-    GetWindowTreeItem(aParent, getter_AddRefs(parentItem));
-
-    if (!callerItem) {
-      callerItem = parentItem;
-    }
-
-    if (parentItem) {
-      parentItem->FindItemWithName(name.get(), nsnull, callerItem,
-                                   getter_AddRefs(newDocShellItem));
-    } else
-      FindItemWithName(name.get(), nsnull, callerItem,
-                       getter_AddRefs(newDocShellItem));
-  }
+  nsCOMPtr<nsIDOMWindow> foundWindow;
+  SafeGetWindowByName(name, aParent, getter_AddRefs(foundWindow));
+  GetWindowTreeItem(foundWindow, getter_AddRefs(newDocShellItem));
 
   // no extant window? make a new one.
 
@@ -580,6 +613,9 @@ nsWindowWatcher::OpenWindowJS(nsIDOMWindow *aParent,
                                      aDialog, uriToLoadIsChrome,
                                      !aParent || chromeParent);
 
+  SizeSpec sizeSpec;
+  CalcSizeSpec(features.get(), sizeSpec);
+
   PRBool isCallerChrome = PR_FALSE;
   nsCOMPtr<nsIScriptSecurityManager>
     sm(do_GetService(NS_SCRIPTSECURITYMANAGER_CONTRACTID));
@@ -597,8 +633,49 @@ nsWindowWatcher::OpenWindowJS(nsIDOMWindow *aParent,
     callerContextGuard.Push(cx);
   }
 
+  if (!newDocShellItem) {
+    // We're going to either open up a new window ourselves or ask a
+    // nsIWindowProvider for one.  In either case, we'll want to set the right
+    // name on it.
+    windowNeedsName = PR_TRUE;
+
+    // Now check whether it's ok to ask a window provider for a window.  Don't
+    // do it if we're opening a dialog or if our parent is a chrome window or
+    // if we're opening something that has modal, dialog, or chrome flags set.
+    nsCOMPtr<nsIDOMChromeWindow> chromeWin = do_QueryInterface(aParent);
+    if (!aDialog && !chromeWin &&
+        !(chromeFlags & (nsIWebBrowserChrome::CHROME_MODAL         |
+                         nsIWebBrowserChrome::CHROME_OPENAS_DIALOG | 
+                         nsIWebBrowserChrome::CHROME_OPENAS_CHROME))) {
+      nsCOMPtr<nsIWindowProvider> provider = do_GetInterface(parentTreeOwner);
+      if (provider) {
+        NS_ASSERTION(aParent, "We've _got_ to have a parent here!");
+
+        nsCOMPtr<nsIDOMWindow> newWindow;
+        rv = provider->ProvideWindow(aParent, chromeFlags,
+                                     sizeSpec.PositionSpecified(),
+                                     sizeSpec.SizeSpecified(),
+                                     uriToLoad, name, features, &windowIsNew,
+                                     getter_AddRefs(newWindow));
+        if (NS_SUCCEEDED(rv)) {
+          GetWindowTreeItem(newWindow, getter_AddRefs(newDocShellItem));
+          if (windowIsNew && newDocShellItem) {
+            // Make sure to stop any loads happening in this window that the
+            // window provider might have started.  Otherwise if our caller
+            // manipulates the window it just opened and then the load
+            // completes their stuff will get blown away.
+            nsCOMPtr<nsIWebNavigation> webNav =
+              do_QueryInterface(newDocShellItem);
+            webNav->Stop(nsIWebNavigation::STOP_NETWORK);
+          }
+        }
+      }
+    }
+  }
+  
   if (!newDocShellItem) {
     windowIsNew = PR_TRUE;
+    isNewToplevelWindow = PR_TRUE;
 
     nsCOMPtr<nsIWebBrowserChrome> parentChrome(do_GetInterface(parentTreeOwner));
 
@@ -682,7 +759,10 @@ nsWindowWatcher::OpenWindowJS(nsIDOMWindow *aParent,
   if (!newDocShellItem)
     return rv;
 
-  rv = ReadyOpenedDocShellItem(newDocShellItem, aParent, _retval);
+  nsCOMPtr<nsIDocShell> newDocShell(do_QueryInterface(newDocShellItem));
+  NS_ENSURE_TRUE(newDocShell, NS_ERROR_UNEXPECTED);
+  
+  rv = ReadyOpenedDocShellItem(newDocShellItem, aParent, windowIsNew, _retval);
   if (NS_FAILED(rv))
     return rv;
 
@@ -692,7 +772,7 @@ nsWindowWatcher::OpenWindowJS(nsIDOMWindow *aParent,
      in the open call as disabling persistence of those attributes.
      Popup windows (which should not persist size or position) generally set
      the size. */
-  if (windowIsNew) {
+  if (isNewToplevelWindow) {
     /* at the moment, the strings "height=" or "width=" never happen
        outside a size specification, so we can do this the Q&D way. */
 
@@ -713,39 +793,108 @@ nsWindowWatcher::OpenWindowJS(nsIDOMWindow *aParent,
     }
   }
 
-  /* allow an extant window to keep its name (important for cases like
-     _self where the given name is different (and invalid)). also _blank
-     is not a window name. */
-  if (windowIsNew)
-    newDocShellItem->SetName(nameSpecified && !name.LowerCaseEqualsLiteral("_blank") ?
+  /* allow a window that we found by name to keep its name (important for cases
+     like _self where the given name is different (and invalid)).  Also, _blank
+     and _new are not window names. */
+  if (windowNeedsName)
+    newDocShellItem->SetName(nameSpecified &&
+                             !name.LowerCaseEqualsLiteral("_blank") &&
+                             !name.LowerCaseEqualsLiteral("_new") ?
                              name.get() : nsnull);
 
-  nsCOMPtr<nsIDocShell> newDocShell(do_QueryInterface(newDocShellItem));
 
-  // When a new window is opened through JavaScript, we want it to use the
-  // charset of its opener as a fallback in the event the document being loaded 
-  // does not specify a charset. Failing to set this charset is not fatal, so we 
-  // want to continue in the face of errors.
-  nsCOMPtr<nsIScriptGlobalObject> parentSGO(do_QueryInterface(aParent));
-  if (parentSGO) {
-    nsIDocShell *parentDocshell = parentSGO->GetDocShell();
-    // parentDocshell may be null if the parent got closed in the meantime
-    if (parentDocshell) {
-      nsCOMPtr<nsIContentViewer> parentContentViewer;
-      parentDocshell->GetContentViewer(getter_AddRefs(parentContentViewer));
-      nsCOMPtr<nsIDocumentViewer> parentDocViewer(do_QueryInterface(parentContentViewer));
-      if (parentDocViewer) {
-        nsCOMPtr<nsIDocument> doc;
-        parentDocViewer->GetDocument(getter_AddRefs(doc));
-        
-        nsCOMPtr<nsIContentViewer> newContentViewer;
-        newDocShell->GetContentViewer(getter_AddRefs(newContentViewer));
-        nsCOMPtr<nsIMarkupDocumentViewer> newMarkupDocViewer(do_QueryInterface(newContentViewer));
-        if (doc && newMarkupDocViewer) {
-          newMarkupDocViewer->SetDefaultCharacterSet(doc->GetDocumentCharacterSet());
+  // Inherit the right character set into the new window to use as a fallback
+  // in the event the document being loaded does not specify a charset.  When
+  // aCalledFromJS is true, we want to use the character set of the document in
+  // the caller; otherwise we want to use the character set of aParent's
+  // docshell. Failing to set this charset is not fatal, so we want to continue
+  // in the face of errors.
+  nsCOMPtr<nsIContentViewer> newCV;
+  newDocShell->GetContentViewer(getter_AddRefs(newCV));
+  nsCOMPtr<nsIMarkupDocumentViewer> newMuCV = do_QueryInterface(newCV);
+  if (newMuCV) {
+    nsCOMPtr<nsIDocShellTreeItem> parentItem;
+    GetWindowTreeItem(aParent, getter_AddRefs(parentItem));
+
+    if (aCalledFromJS) {
+      nsCOMPtr<nsIDocShellTreeItem> callerItem = GetCallerTreeItem(parentItem);
+      nsCOMPtr<nsPIDOMWindow> callerWin = do_GetInterface(callerItem);
+      if (callerWin) {
+        nsCOMPtr<nsIDocument> doc =
+          do_QueryInterface(callerWin->GetExtantDocument());
+        if (doc) {
+          newMuCV->SetDefaultCharacterSet(doc->GetDocumentCharacterSet());
         }
       }
     }
+    else {
+      nsCOMPtr<nsIDocShell> parentDocshell = do_QueryInterface(parentItem);
+      // parentDocshell may be null if the parent got closed in the meantime
+      if (parentDocshell) {
+        nsCOMPtr<nsIContentViewer> parentCV;
+        parentDocshell->GetContentViewer(getter_AddRefs(parentCV));
+        nsCOMPtr<nsIMarkupDocumentViewer> parentMuCV =
+          do_QueryInterface(parentCV);
+        if (parentMuCV) {
+          nsCAutoString charset;
+          nsresult res = parentMuCV->GetDefaultCharacterSet(charset);
+          if (NS_SUCCEEDED(res)) {
+            newMuCV->SetDefaultCharacterSet(charset);
+          }
+          res = parentMuCV->GetPrevDocCharacterSet(charset);
+          if (NS_SUCCEEDED(res)) {
+            newMuCV->SetPrevDocCharacterSet(charset);
+          }
+        }
+      }
+    }
+  }
+
+  if (isNewToplevelWindow) {
+    // Notify observers that the window is open and ready.
+    // The window has not yet started to load a document.
+    nsCOMPtr<nsIObserverService> obsSvc =
+      do_GetService("@mozilla.org/observer-service;1");
+    if (obsSvc) {
+      obsSvc->NotifyObservers(*_retval, "toplevel-window-ready", nsnull);
+    }
+  }
+
+  // Now we have to set the right opener principal on the new window.  Note
+  // that we have to do this _before_ starting any URI loads, thanks to the
+  // sync nature of javascript: loads.  Since this is the only place where we
+  // set said opener principal, we need to do it for all URIs, including
+  // chrome ones.  So to deal with the mess that is bug 79775, just press on in
+  // a reasonable way even if GetSubjectPrincipal fails.  In that case, just
+  // use a null subjectPrincipal.
+  nsCOMPtr<nsIPrincipal> subjectPrincipal;
+  if (NS_FAILED(sm->GetSubjectPrincipal(getter_AddRefs(subjectPrincipal)))) {
+    subjectPrincipal = nsnull;
+  }
+
+  if (windowIsNew) {
+    // Now set the opener principal on the new window.  Note that we need to do
+    // this no matter whether we were opened from JS; if there is nothing on
+    // the JS stack, just use the principal of our parent window.  In those
+    // cases we do _not_ set the parent window principal as the owner of the
+    // load--since we really don't know who the owner is, just leave it null.
+    nsIPrincipal* newWindowPrincipal = subjectPrincipal;
+    if (!newWindowPrincipal && aParent) {
+      nsCOMPtr<nsIScriptObjectPrincipal> sop(do_QueryInterface(aParent));
+      if (sop) {
+        newWindowPrincipal = sop->GetPrincipal();
+      }
+    }
+
+    nsCOMPtr<nsPIDOMWindow_MOZILLA_1_8_BRANCH2> newWindow =
+      do_QueryInterface(*_retval);
+#ifdef DEBUG
+    nsCOMPtr<nsPIDOMWindow> newDebugWindow = do_GetInterface(newDocShell);
+    NS_ASSERTION(newWindow == newDebugWindow, "Different windows??");
+#endif
+    if (newWindow) {
+      newWindow->SetOpenerScriptPrincipal(newWindowPrincipal);
+    }
   }
 
   if (uriToLoad) { // get the script principal and pass it to docshell
@@ -767,15 +916,8 @@ nsWindowWatcher::OpenWindowJS(nsIDOMWindow *aParent,
     newDocShell->CreateLoadInfo(getter_AddRefs(loadInfo));
     NS_ENSURE_TRUE(loadInfo, NS_ERROR_FAILURE);
 
-    if (!uriToLoadIsChrome) {
-      nsCOMPtr<nsIPrincipal> principal;
-      if (NS_FAILED(sm->GetSubjectPrincipal(getter_AddRefs(principal))))
-        return NS_ERROR_FAILURE;
-
-      if (principal) {
-        nsCOMPtr<nsISupports> owner(do_QueryInterface(principal));
-        loadInfo->SetOwner(owner);
-      }
+    if (subjectPrincipal) {
+      loadInfo->SetOwner(subjectPrincipal);
     }
 
     // Set the new window's referrer from the calling context's document:
@@ -787,8 +929,7 @@ nsWindowWatcher::OpenWindowJS(nsIDOMWindow *aParent,
 
     // get its document, if any
     if (stack && NS_SUCCEEDED(stack->Peek(&ccx)) && ccx) {
-      nsIScriptGlobalObject *sgo =
-        nsWWJSUtils::GetStaticScriptGlobal(ccx, ::JS_GetGlobalObject(ccx));
+      nsIScriptGlobalObject *sgo = nsWWJSUtils::GetDynamicScriptGlobal(ccx);
 
       nsCOMPtr<nsPIDOMWindow> w(do_QueryInterface(sgo));
       if (w) {
@@ -807,12 +948,12 @@ nsWindowWatcher::OpenWindowJS(nsIDOMWindow *aParent,
     }
 
     newDocShell->LoadURI(uriToLoad, loadInfo,
-                         nsIWebNavigation::LOAD_FLAGS_NONE, PR_TRUE);
+      windowIsNew ? nsIWebNavigation::LOAD_FLAGS_FIRST_LOAD :
+                    nsIWebNavigation::LOAD_FLAGS_NONE, PR_TRUE);
   }
 
-  if (windowIsNew)
-    SizeOpenedDocShellItem(newDocShellItem, aParent, features.get(),
-                           chromeFlags);
+  if (isNewToplevelWindow)
+    SizeOpenedDocShellItem(newDocShellItem, aParent, sizeSpec);
 
   if (windowIsModal) {
     nsCOMPtr<nsIDocShellTreeOwner> newTreeOwner;
@@ -1017,7 +1158,7 @@ nsWindowWatcher::FindWindowEntry(nsIDOMWindow *aWindow)
   found = PR_FALSE;
   while (info != listEnd && NS_SUCCEEDED(rv)) {
     nsCOMPtr<nsIDOMWindow> infoWindow(do_QueryReferent(info->mWindow));
-    if (!infoWindow) // clean up dangling reference, while we're here
+    if (!infoWindow) { // clean up dangling reference, while we're here
       rv = RemoveWindow(info);
     }
     else if (infoWindow.get() == aWindow)
@@ -1111,37 +1252,22 @@ nsWindowWatcher::GetWindowByName(const PRUnichar *aTargetName,
 
   *aResult = nsnull;
 
-  nsCOMPtr<nsIWebNavigation> webNav;
   nsCOMPtr<nsIDocShellTreeItem> treeItem;
 
-  // First, check if the TargetName exists in the aCurrentWindow hierarchy
-  webNav = do_GetInterface(aCurrentWindow);
-  if (webNav) {
-    nsCOMPtr<nsIDocShellTreeItem> docShellTreeItem;
-
-    docShellTreeItem = do_QueryInterface(webNav);
-    if (docShellTreeItem) {
-      // Note: original requestor is null here, per idl comments
-      docShellTreeItem->FindItemWithName(aTargetName, nsnull, nsnull,
-                                         getter_AddRefs(treeItem));
-    }
+  nsCOMPtr<nsIDocShellTreeItem> startItem;
+  GetWindowTreeItem(aCurrentWindow, getter_AddRefs(startItem));
+  if (startItem) {
+    // Note: original requestor is null here, per idl comments
+    startItem->FindItemWithName(aTargetName, nsnull, nsnull,
+                                getter_AddRefs(treeItem));
   }
-
-  // Next, see if the TargetName exists in any window hierarchy
-  if (!treeItem) {
+  else {
     // Note: original requestor is null here, per idl comments
     FindItemWithName(aTargetName, nsnull, nsnull, getter_AddRefs(treeItem));
   }
 
-  if (treeItem) {
-    nsCOMPtr<nsIDOMWindow> domWindow;
-
-    domWindow = do_GetInterface(treeItem);
-    if (domWindow) {
-      *aResult = domWindow;
-      NS_ADDREF(*aResult);
-    }
-  }
+  nsCOMPtr<nsIDOMWindow> domWindow = do_GetInterface(treeItem);
+  domWindow.swap(*aResult);
 
   return NS_OK;
 }
@@ -1203,6 +1329,7 @@ nsWindowWatcher::URIfromURL(const char *aURL,
   return NS_NewURI(aURI, aURL, baseURI);
 }
 
+#ifdef DEBUG
 /* Check for an illegal name e.g. frame3.1
    This just prints a warning message an continues; we open the window anyway,
    (see bug 32898). */
@@ -1218,15 +1345,14 @@ void nsWindowWatcher::CheckWindowName(nsString& aName)
 
       // Don't use js_ReportError as this will cause the application
       // to shut down (JS_ASSERT calls abort())  See bug 32898
-      nsAutoString warn;
+      nsCAutoString warn;
       warn.AssignLiteral("Illegal character in window name ");
-      warn.Append(aName);
-      char *cp = ToNewCString(warn);
-      NS_WARNING(cp);
-      nsCRT::free(cp);
+      AppendUTF16toUTF8(aName, warn);
+      NS_WARNING(warn.get());
       break;
     }
 }
+#endif // DEBUG
 
 #define NS_CALCULATE_CHROME_FLAG_FOR(feature, flag)               \
     prefBranch->GetBoolPref(feature, &forceEnable);               \
@@ -1395,6 +1521,11 @@ PRUint32 nsWindowWatcher::CalculateChromeFlags(const char *aFeatures,
       chromeFlags &= ~(nsIWebBrowserChrome::CHROME_MODAL | nsIWebBrowserChrome::CHROME_OPENAS_CHROME);
   }
 
+  if (!(chromeFlags & nsIWebBrowserChrome::CHROME_OPENAS_CHROME)) {
+    // Remove the dependent flag if we're not opening as chrome
+    chromeFlags &= ~nsIWebBrowserChrome::CHROME_DEPENDENT;
+  }
+
   return chromeFlags;
 }
 
@@ -1513,6 +1644,65 @@ nsWindowWatcher::FindItemWithName(const PRUnichar* aName,
   return rv;
 }
 
+already_AddRefed<nsIDocShellTreeItem>
+nsWindowWatcher::GetCallerTreeItem(nsIDocShellTreeItem* aParentItem)
+{
+  nsCOMPtr<nsIJSContextStack> stack =
+    do_GetService(sJSStackContractID);
+
+  JSContext *cx = nsnull;
+
+  if (stack) {
+    stack->Peek(&cx);
+  }
+
+  nsIDocShellTreeItem* callerItem = nsnull;
+
+  if (cx) {
+    nsCOMPtr<nsIWebNavigation> callerWebNav =
+      do_GetInterface(nsWWJSUtils::GetDynamicScriptGlobal(cx));
+
+    if (callerWebNav) {
+      CallQueryInterface(callerWebNav, &callerItem);
+    }
+  }
+
+  if (!callerItem) {
+    NS_IF_ADDREF(callerItem = aParentItem);
+  }
+
+  return callerItem;
+}
+
+nsresult
+nsWindowWatcher::SafeGetWindowByName(const nsAString& aName,
+                                     nsIDOMWindow* aCurrentWindow,
+                                     nsIDOMWindow** aResult)
+{
+  *aResult = nsnull;
+  
+  nsCOMPtr<nsIDocShellTreeItem> startItem;
+  GetWindowTreeItem(aCurrentWindow, getter_AddRefs(startItem));
+
+  nsCOMPtr<nsIDocShellTreeItem> callerItem = GetCallerTreeItem(startItem);
+
+  const nsAFlatString& flatName = PromiseFlatString(aName);
+
+  nsCOMPtr<nsIDocShellTreeItem> foundItem;
+  if (startItem) {
+    startItem->FindItemWithName(flatName.get(), nsnull, callerItem,
+                                getter_AddRefs(foundItem));
+  }
+  else {
+    FindItemWithName(flatName.get(), nsnull, callerItem,
+                     getter_AddRefs(foundItem));
+  }
+
+  nsCOMPtr<nsIDOMWindow> foundWin = do_GetInterface(foundItem);
+  foundWin.swap(*aResult);
+  return NS_OK;
+}
+
 /* Fetch the nsIDOMWindow corresponding to the given nsIDocShellTreeItem.
    This forces the creation of a script context, if one has not already
    been created. Note it also sets the window's opener to the parent,
@@ -1521,6 +1711,7 @@ nsWindowWatcher::FindItemWithName(const PRUnichar* aName,
 nsresult
 nsWindowWatcher::ReadyOpenedDocShellItem(nsIDocShellTreeItem *aOpenedItem,
                                          nsIDOMWindow        *aParent,
+                                         PRBool              aWindowIsNew,
                                          nsIDOMWindow        **aOpenedWindow)
 {
   nsresult rv = NS_ERROR_FAILURE;
@@ -1530,14 +1721,86 @@ nsWindowWatcher::ReadyOpenedDocShellItem(nsIDocShellTreeItem *aOpenedItem,
   if (globalObject) {
     if (aParent) {
       nsCOMPtr<nsIDOMWindowInternal> internalParent(do_QueryInterface(aParent));
-      globalObject->SetOpenerWindow(internalParent); // damnit
+      nsCOMPtr<nsPIDOMWindow_MOZILLA_1_8_BRANCH> branchWindow =
+        do_QueryInterface(globalObject);
+      branchWindow->SetOpenerWindow(internalParent, aWindowIsNew); // damnit
+
+      if (aWindowIsNew) {
+        nsCOMPtr<nsIDocument_MOZILLA_1_8_BRANCH2> doc =
+          do_QueryInterface(branchWindow->GetExtantDocument());
+        if (doc) {
+          doc->SetIsInitialDocument(PR_TRUE);
+        }
+      }
     }
     rv = CallQueryInterface(globalObject, aOpenedWindow);
   }
   return rv;
 }
 
-/* Size and position the new window according to aFeatures. This method
+void
+nsWindowWatcher::CalcSizeSpec(const char* aFeatures, SizeSpec& aResult)
+{
+  // Parse position spec, if any, from aFeatures
+  PRBool  present;
+  PRInt32 temp;
+
+  present = PR_FALSE;
+  if ((temp = WinHasOption(aFeatures, "left", 0, &present)) || present)
+    aResult.mLeft = temp;
+  else if ((temp = WinHasOption(aFeatures, "screenX", 0, &present)) || present)
+    aResult.mLeft = temp;
+  aResult.mLeftSpecified = present;
+
+  present = PR_FALSE;
+  if ((temp = WinHasOption(aFeatures, "top", 0, &present)) || present)
+    aResult.mTop = temp;
+  else if ((temp = WinHasOption(aFeatures, "screenY", 0, &present)) || present)
+    aResult.mTop = temp;
+  aResult.mTopSpecified = present;
+
+  // Parse size spec, if any. Chrome size overrides content size.
+  if ((temp = WinHasOption(aFeatures, "outerWidth", PR_INT32_MIN, nsnull))) {
+    if (temp == PR_INT32_MIN) {
+      aResult.mUseDefaultWidth = PR_TRUE;
+    }
+    else {
+      aResult.mOuterWidth = temp;
+    }
+    aResult.mOuterWidthSpecified = PR_TRUE;
+  } else if ((temp = WinHasOption(aFeatures, "width", PR_INT32_MIN, nsnull)) ||
+             (temp = WinHasOption(aFeatures, "innerWidth", PR_INT32_MIN,
+                                  nsnull))) {
+    if (temp == PR_INT32_MIN) {
+      aResult.mUseDefaultWidth = PR_TRUE;
+    } else {
+      aResult.mInnerWidth = temp;
+    }
+    aResult.mInnerWidthSpecified = PR_TRUE;
+  }
+
+  if ((temp = WinHasOption(aFeatures, "outerHeight", PR_INT32_MIN, nsnull))) {
+    if (temp == PR_INT32_MIN) {
+      aResult.mUseDefaultHeight = PR_TRUE;
+    }
+    else {
+      aResult.mOuterHeight = temp;
+    }
+    aResult.mOuterHeightSpecified = PR_TRUE;
+  } else if ((temp = WinHasOption(aFeatures, "height", PR_INT32_MIN,
+                                  nsnull)) ||
+             (temp = WinHasOption(aFeatures, "innerHeight", PR_INT32_MIN,
+                                  nsnull))) {
+    if (temp == PR_INT32_MIN) {
+      aResult.mUseDefaultHeight = PR_TRUE;
+    } else {
+      aResult.mInnerHeight = temp;
+    }
+    aResult.mInnerHeightSpecified = PR_TRUE;
+  }
+}
+
+/* Size and position the new window according to aSizeSpec. This method
    is assumed to be called after the window has already been given
    a default position and size; thus its current position and size are
    accurate defaults. The new window is made visible at method end.
@@ -1545,8 +1808,7 @@ nsWindowWatcher::ReadyOpenedDocShellItem(nsIDocShellTreeItem *aOpenedItem,
 void
 nsWindowWatcher::SizeOpenedDocShellItem(nsIDocShellTreeItem *aDocShellItem,
                                         nsIDOMWindow *aParent,
-                                        const char *aFeatures,
-                                        PRUint32 aChromeFlags)
+                                        const SizeSpec & aSizeSpec)
 {
   // position and size of window
   PRInt32 left = 0,
@@ -1585,62 +1847,47 @@ nsWindowWatcher::SizeOpenedDocShellItem(nsIDocShellTreeItem *aDocShellItem,
     }
   }
 
-  // Parse position spec, if any, from aFeatures
-
-  PRBool  positionSpecified = PR_FALSE;
-  PRBool  present;
-  PRInt32 temp;
-
-  present = PR_FALSE;
-  if ((temp = WinHasOption(aFeatures, "left", 0, &present)) || present)
-    left = temp;
-  else if ((temp = WinHasOption(aFeatures, "screenX", 0, &present)) || present)
-    left = temp;
-  if (present)
-    positionSpecified = PR_TRUE;
-
-  present = PR_FALSE;
-  if ((temp = WinHasOption(aFeatures, "top", 0, &present)) || present)
-    top = temp;
-  else if ((temp = WinHasOption(aFeatures, "screenY", 0, &present)) || present)
-    top = temp;
-  if (present)
-    positionSpecified = PR_TRUE;
-
-  PRBool sizeSpecified = PR_FALSE;
-
-  // Parse size spec, if any. Chrome size overrides content size.
-
-  if ((temp = WinHasOption(aFeatures, "outerWidth", width, nsnull))) {
-    width = temp;
-    sizeSpecified = PR_TRUE;
-  } else if ((temp = WinHasOption(aFeatures, "width",
-                                  width - chromeWidth, nsnull))) {
-    width = temp;
-    sizeChromeWidth = PR_FALSE;
-    sizeSpecified = PR_TRUE;
-  } else if ((temp = WinHasOption(aFeatures, "innerWidth",
-                                  width - chromeWidth, nsnull))) {
-    width = temp;
-    sizeChromeWidth = PR_FALSE;
-    sizeSpecified = PR_TRUE;
+  // Set up left/top
+  if (aSizeSpec.mLeftSpecified) {
+    left = aSizeSpec.mLeft;
   }
 
-  if ((temp = WinHasOption(aFeatures, "outerHeight", height, nsnull))) {
-    height = temp;
-    sizeSpecified = PR_TRUE;
-  } else if ((temp = WinHasOption(aFeatures, "height",
-                                  height - chromeHeight, nsnull))) {
-    height = temp;
-    sizeChromeHeight = PR_FALSE;
-    sizeSpecified = PR_TRUE;
-  } else if ((temp = WinHasOption(aFeatures, "innerHeight",
-                                  height - chromeHeight, nsnull))) {
-    height = temp;
-    sizeChromeHeight = PR_FALSE;
-    sizeSpecified = PR_TRUE;
+  if (aSizeSpec.mTopSpecified) {
+    top = aSizeSpec.mTop;
   }
 
+  // Set up width
+  if (aSizeSpec.mOuterWidthSpecified) {
+    if (!aSizeSpec.mUseDefaultWidth) {
+      width = aSizeSpec.mOuterWidth;
+    } // Else specified to default; just use our existing width
+  }
+  else if (aSizeSpec.mInnerWidthSpecified) {
+    sizeChromeWidth = PR_FALSE;
+    if (aSizeSpec.mUseDefaultWidth) {
+      width = width - chromeWidth;
+    } else {
+      width = aSizeSpec.mInnerWidth;
+    }
+  }
+
+  // Set up height
+  if (aSizeSpec.mOuterHeightSpecified) {
+    if (!aSizeSpec.mUseDefaultHeight) {
+      height = aSizeSpec.mOuterHeight;
+    } // Else specified to default; just use our existing height
+  }
+  else if (aSizeSpec.mInnerHeightSpecified) {
+    sizeChromeHeight = PR_FALSE;
+    if (aSizeSpec.mUseDefaultHeight) {
+      height = height - chromeHeight;
+    } else {
+      height = aSizeSpec.mInnerHeight;
+    }
+  }
+
+  PRBool positionSpecified = aSizeSpec.PositionSpecified();
+  
   nsresult res;
   PRBool enabled = PR_FALSE;
 
@@ -1686,7 +1933,7 @@ nsWindowWatcher::SizeOpenedDocShellItem(nsIDocShellTreeItem *aDocShellItem,
       screen->GetAvailRect(&screenLeft, &screenTop,
                            &screenWidth, &screenHeight);
 
-      if (sizeSpecified) {
+      if (aSizeSpec.SizeSpecified()) {
         /* Unlike position, force size out-of-bounds check only if
            size actually was specified. Otherwise, intrinsically sized
            windows are broken. */
@@ -1717,7 +1964,7 @@ nsWindowWatcher::SizeOpenedDocShellItem(nsIDocShellTreeItem *aDocShellItem,
 
   if (positionSpecified)
     treeOwnerAsWin->SetPosition(left, top);
-  if (sizeSpecified) {
+  if (aSizeSpec.SizeSpecified()) {
     /* Prefer to trust the interfaces, which think in terms of pure
        chrome or content sizes. If we have a mix, use the chrome size
        adjusted by the chrome/content differences calculated earlier. */
@@ -1755,7 +2002,8 @@ nsWindowWatcher::ConvertSupportsTojsvals(nsIDOMWindow *aWindow,
                                          nsISupports *aArgs,
                                          PRUint32 *aArgc, jsval **aArgv,
                                          JSContext **aUsedContext,
-                                         void **aMarkp)
+                                         void **aMarkp,
+                                         nsIScriptContext **aScriptContext)
 {
   nsresult rv = NS_OK;
 
@@ -1782,6 +2030,15 @@ nsWindowWatcher::ConvertSupportsTojsvals(nsIDOMWindow *aWindow,
   JSContextAutoPopper  contextGuard;
 
   cx = GetJSContextFromWindow(aWindow);
+  if (cx) {
+    // Our caller needs to hold a strong ref to keep this context alive.
+    *aScriptContext = GetScriptContextFromJSContext(cx);
+    NS_ASSERTION(*aScriptContext,
+                 "The window's context doesn't have a script context?");
+    NS_ADDREF(*aScriptContext);
+  } else {
+    *aScriptContext = nsnull;
+  }
   if (!cx)
     cx = GetJSContextFromCallStack();
   if (!cx) {
@@ -2103,7 +2360,7 @@ nsWindowWatcher::GetJSContextFromWindow(nsIDOMWindow *aWindow)
     }
     /* (off-topic note:) the nsIScriptContext can be retrieved by
     nsCOMPtr<nsIScriptContext> scx;
-    nsJSUtils::nsGetDynamicScriptContext(cx, getter_AddRefs(scx));
+    nsJSUtils::GetDynamicScriptContext(cx, getter_AddRefs(scx));
     */
   }
 
